unit TarModule;

interface

uses
	Windows, SysUtils, Math,
	UnifyArc, Tar32,
	{$IFNDEF EXTENSION}LObjects, {$ENDIF} EOFileIntf, ArchiversShared;

type
	TFileArray = array of TFile;

	TArchiveType = (atOthers, atTar, atTarGZ, atTarZ, atGZ, atZ, atTarBZ2, atBZ2);

	TTarArchive = class(TInterfacedObject, IFolder, IArchive)
	private
		FPath: TPath;
		FFiles: TFileArray;
		FViews: array of IObserver;
		FValid: Boolean;
		FType: TArchiveType;
		{IObservable}
		procedure Attach(const AView: IObserver);
		procedure Detach(const AView: IObserver);
		function HasClients: Boolean;
		{IFolder}
		function GetPath: TPath;
		function GetItems(AIndex: Integer): PFile;
		function GetCount: Integer;
		function GetValid: Boolean;
		function GetDisplayText: WideString;
		procedure Refresh;
		procedure Move(const NewPath: WideString);
		function GetParentPath: TPath;
		function IsSub(const AParent: TPath): Boolean;
		procedure StopSurveillance;
		procedure ResumeSurveillance(Wait: Boolean);
		function GetInfoText(AFile: PFile): WideString;
		procedure UnpackMemory(AFile: PFile; var Data: String);
		procedure Unpack(AFile: PFile; const DestDir: WideString);
		function GetPossibilities: TFolderPossibilities;
		function ConcatenatePath(AFile: PFile): TPath;
	protected
		procedure Notify;
		procedure DoRefresh;
	public
		constructor Create(const APath: TPath);
	end;

	TTarArchiver = class(TInterfacedObject, IExtension, IExtensionInfo, IArchiver)
	public
		{IExtension}
		procedure Initialize;
		procedure Finalize;
		{IExtensionInfo}
		function GetName: WideString;
		procedure Dialog;
		{IArchiver}
		function Enumerate(Proc: CEachArchiveKind; Context: Pointer): WideString;
		function Open(const APath: TPath): IFolder;
		procedure Pack(const Archive, SourceDir: WideString; const Sources: array of WideString);
		procedure Unpack(const Archive, DestDir: WideString);
	end;

implementation

function MakeResponseFile(const BasePath: string; const Files: array of WideString): TFileName;
var
	H: THandle;
	F, Line: string;
	I: Integer;
begin
	Result := TemporaryPath + '$$$$$$$$.lst';

	H := FileCreate(Result);
	if H = INVALID_HANDLE_VALUE then
		raise EInOutError.CreateFmt('X|Xt@C%s쐬ł܂ł.', [Result]);

	try
		for I := Low(Files) to High(Files) do
		begin
			F := Files[I];

			if DirectoryExists(BasePath + F) then
				Line := '"' + F + '\*"'#13#10
			else
				Line := '"' + F + '"'#13#10;

			FileWrite(H, PChar(Line)^, Length(Line));
		end;
	finally
		FileClose(H);
	end;
end;

function ExtractArchiveDir(const S: WideString): WideString;
begin
	Result := S;
	Delete(Result, Max(1, WideLastDelimiter('/', Result)), MaxInt);
end;

{ TTarArchive }

procedure TTarArchive.Attach(const AView: IObserver);
begin
	SetLength(FViews, Length(FViews) + 1);
	FViews[High(FViews)] := AView;
end;

function TTarArchive.ConcatenatePath(AFile: PFile): TPath;
var
	S: WideString;
begin
	Result := EmptyPath;

	Result.Directory := FPath.Directory;
	Result.Archive := FPath.Archive;

	with AFile^ do SetString(S, Name, NameLength);

	if Result.Archive <> '' then
		Result.Archive := Result.Archive + '/';

	Result.Archive := Result.Archive + S;
end;

constructor TTarArchive.Create(const APath: TPath);
begin
	inherited Create;
	FPath := APath;

	SetCursorWait;
	DoRefresh;
	ResetCursor;
end;

procedure TTarArchive.Detach(const AView: IObserver);
var
	I, J: Integer;
begin
	for I := High(FViews) downto Low(FViews) do
		if FViews[I] = AView then
		begin
			for J := I + 1 to High(FViews) do
				FViews[J - 1] := FViews[J];

			SetLength(FViews, Length(FViews) - 1);
			Break;
		end;
end;

procedure TTarArchive.DoRefresh;
label
	Next;
const
	BufferFirst = 1024;
	BufferStep = 256;
var
	Info: TINDIVIDUALINFO;
	Sub, I, Count: Integer;
	H: THandle;
	FileName: WideString;
	ArcNameA, PathInArc: string;
	F: TFile;
	PFileName: PChar;
begin
	FValid := False;

	Count := 0;
	SetLength(FFiles, BufferFirst);

	ArcNameA := string(FPath.Directory);

	case TarGetArchiveType(PChar(ArcNameA)) of
	ARCHIVETYPE_TAR:    FType := atTar;
	ARCHIVETYPE_TARGZ:  FType := atTarGZ;
	ARCHIVETYPE_TARZ:   FType := atTarZ;
	ARCHIVETYPE_GZ:     FType := atGZ;
	ARCHIVETYPE_Z:      FType := atZ;
	ARCHIVETYPE_TARBZ2: FType := atTarBZ2;
	ARCHIVETYPE_BZ2:    FType := atBZ2;
	else                FType := atOthers;
	end;

	H := TarOpenArchive(0, PChar(ArcNameA),
					M_REGARDLESS_INIT_FILE or M_NOT_USE_DRIVE_LETTER or
					M_CHECK_FILENAME_ONLY or M_ERROR_MESSAGE_ON or M_NO_BACKGROUND_MODE);
	if H = 0 then	Exit;

	PathInArc := '';
	if FPath.Archive <> '' then
		PathInArc := StringReplace(FPath.Archive, '\', '/', [rfReplaceAll]) + '/';

	if TarFindFirst(H, '*', Info) = 0 then
		repeat
			PFileName := Info.szFileName;

			if StrLIComp(PFileName, PChar(PathInArc), Length(PathInArc)) <> 0 then
				Continue;

			Inc(PFileName, Length(PathInArc));
			if PFileName^ = #0 then
				Continue;

			FileName := StringReplace(PFileName, '\', '/', [rfReplaceAll]);
			if FileName[Length(FileName)] = '/' then
				Delete(FileName, Length(FileName), 1);

			Sub := Pos('/', FileName);
			if Sub > 0 then
			begin
				FileName := Copy(FileName, 1, Sub - 1);

				for I := Count - 1 downto 0 do
					if lstrcmpW(FFiles[I].Name, PWideChar(FileName)) = 0 then
						goto Next;

				F.Kind := fkFile;

				lstrcpyW(F.Name, PWideChar(FileName));
				F.NameLength := Length(FileName);
				F.DisplayName := F.Name;
				F.DisplayNameLength := F.NameLength;
				F.ShortNameLength := 0;

				F.CreationTime.dwLowDateTime := 0;
				F.CreationTime.dwHighDateTime := 0;
				F.LastAccessTime.dwLowDateTime := 0;
				F.LastAccessTime.dwHighDateTime := 0;
				F.LastWriteTime.dwLowDateTime := 0;
				F.LastWriteTime.dwHighDateTime := 0;

				F.Size := 0;
				F.Attributes := faDirectory;

				F.Operatable := [foEnter];
			end
			else
			begin
				F.Kind := fkFile;

				lstrcpyW(F.Name, PWideChar(FileName));
				F.NameLength := Length(FileName);
				F.DisplayName := F.Name;
				F.DisplayNameLength := F.NameLength;
				F.ShortNameLength := 0;

				F.CreationTime := UnixUTCToFileTime(TarGetCreateTime(H));
				F.LastAccessTime := UnixUTCToFileTime(TarGetAccessTime(H));
				F.LastWriteTime := UnixUTCToFileTime(TarGetWriteTime(H));

				F.Size := Info.dwOriginalSize;
				F.Attributes := TarGetAttribute(H);

				F.Operatable := [];
				if F.Attributes and faDirectory <> 0 then
					Include(F.Operatable, foEnter);
			end;
			F.Number := High(Word);

			{Add}
			Inc(Count);
			if Count > Length(FFiles) then
				SetLength(FFiles, Length(FFiles) + BufferStep);
			FFiles[Count - 1] := F;
Next:
		until TarFindNext(H, Info) <> 0;
	TarCloseArchive(H);

	SetLength(FFiles, Count);
	FValid := True;
end;

function TTarArchive.GetCount: Integer;
begin
	Result := Length(FFiles)
end;

function TTarArchive.GetDisplayText: WideString;
begin
	Result := FPath.Directory;
	if FPath.Archive <> '' then
		Result := Result + '/' + FPath.Archive
end;

function TTarArchive.GetInfoText(AFile: PFile): WideString;
const
	T: array[TArchiveType] of WideString = (
		'in Others', 'in Tar', 'in Tar.GZ', 'in Tar.Z',
		'in GZ', 'in Z', 'in Tar.BZ2', 'in BZ2');
begin
	Result := T[FType]
end;

function TTarArchive.GetItems(AIndex: Integer): PFile;
begin
	Result := @FFiles[AIndex]
end;

function TTarArchive.GetParentPath: TPath;
begin
	Result := FPath;

	if Result.Archive = '' then
		Result.Directory := WideExtractFileDir(Result.Directory)
	else
		Result.Archive := ExtractArchiveDir(Result.Archive)
end;

function TTarArchive.GetPath: TPath;
begin
	Result := FPath
end;

function TTarArchive.GetPossibilities: TFolderPossibilities;
begin
	Result := [fpSort, fpHasParent]
end;

function TTarArchive.GetValid: Boolean;
begin
	Result := FValid
end;

function TTarArchive.HasClients: Boolean;
begin
	Result := FViews <> nil
end;

function TTarArchive.IsSub(const AParent: TPath): Boolean;
var
	ParentDir, ParentArc: WideString;
	ParentNameSpace: INameSpace;
begin
	if AParent.ItemIDList = nil then
	begin
		if AParent.Directory = FPath.Directory then
		begin
			ParentArc := AParent.Archive + '/';
			Result := WideCompareFileName(ParentArc, Copy(FPath.Archive + '/', 1, Length(ParentArc))) = 0
		end
		else
		begin
			ParentDir := WideIncludeTrailingPathDelimiter(AParent.Directory);
			Result := WideCompareFileName(ParentDir, Copy(FPath.Directory + '\', 1, Length(ParentDir))) = 0
		end
	end
	else
	begin
		{fXNgbv}CRs[^}
		ParentNameSpace := (GetApplication as IFolderManager).GetFolder(AParent) as INameSpace;
		Result := ParentNameSpace.AtDesktop or ParentNameSpace.AtDrives;
	end;
end;

procedure TTarArchive.Move(const NewPath: WideString);
begin
	FPath.Directory := NewPath
end;

procedure TTarArchive.Notify;
var
	I: Integer;
	C: Pointer;
begin
(*
	BeginNotification;
	try
*)
		{EĂvȂ悤Ɂc}
		I := Low(FViews);
		while I <= High(FViews) do
		begin
			C := Pointer(FViews[I]);
			FViews[I].Observe;
			if C = Pointer(FViews[I]) then
				Inc(I);
		end;
(*
	finally
		EndNotification
	end
*)
end;

procedure TTarArchive.Refresh;
var
	OldFiles: TFileArray;
begin
	if FValid then
	begin
		SetCursorWait;

		OldFiles := FFiles; {View ܂QƂĂ\邽߉Ȃ}
		try
			DoRefresh;
		except
			FValid := False;
		end;
		ResetCursor;

		Notify; {ʒm}
	end;

	OldFiles := nil; {sage...}
end;

procedure TTarArchive.ResumeSurveillance;
begin
end;

procedure TTarArchive.StopSurveillance;
begin
end;

procedure TTarArchive.Unpack(AFile: PFile; const DestDir: WideString);
var
	DestPath, F, Full: WideString;
	Cmd: string;
	Buffer: array[0..4095] of Char;
begin
	DestPath := WideIncludeTrailingPathDelimiter(DestDir);

	with AFile^ do SetString(F, Name, NameLength);
	if FPath.Archive <> '' then F := FPath.Archive + '/' + F;

	Full := DestPath + F;

	Cmd := '-x --use-directory=0 --display-dialog=0 "' + FPath.Directory + '" "' + DestPath + '" "' + F + '"';
	if Tar(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise ETarError.Create(Buffer);
end;

procedure TTarArchive.UnpackMemory(AFile: PFile; var Data: String);
var
	Temp, PureName, WorkS: WideString;
	PathInArc, Cmd: string;
	Buffer: array[0..4095] of Char;
	H: THandle;
begin
	{tar32.dllExtractMemT|[gȂ̂tempɓWJ}
	WorkS := TemporaryPath;

	with AFile^ do SetString(PureName, Name, NameLength);

	PathInArc := PureName;
	if FPath.Archive <> '' then PathInArc := FPath.Archive + '/' + PureName;

	Cmd := '-x --use-directory=0 "' + FPath.Directory + '" "' + WorkS + '" "' + PathInArc + '"';
	if Tar(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise ETarError.Create(Buffer);

	Temp := WorkS + PureName;

	H := FileOpen(Temp, fmOpenRead);
	if H = INVALID_HANDLE_VALUE then RaiseLastOSError;
	SetLength(Data, GetFileSize(H, nil));
	FileRead(H, Pointer(Data)^, Length(Data));
	FileClose(H);

	DeleteFileW(PWideChar(Temp));
end;

{ TTarArchiver }

procedure TTarArchiver.Dialog;
var
	Buffer: array[0..1023] of Char;
begin
	try
		TarConfigDialog(0, Buffer, UNPACK_CONFIG_MODE)
	except
		MessageBoxW(0, 'tar32.dll܂', 'TAR', MB_ICONERROR)
	end
end;

function TTarArchiver.Enumerate(Proc: CEachArchiveKind; Context: Pointer): WideString;
begin
	if Proc('.tar', Context) then
		Result := '.tar'
	else if Proc('.tgz', Context) then
		Result := '.tgz'
	else if Proc('.gz', Context) then
		Result := '.gz'
	else if Proc('.z', Context) then
		Result := '.z'
	else if Proc('.bz2', Context) then
		Result := '.bz2'
	else
		Result := ''
end;

procedure TTarArchiver.Finalize;
begin
end;

function TTarArchiver.GetName: WideString;
begin
	Result := 'tar32.dll'
end;

procedure TTarArchiver.Initialize;
begin
end;

function TTarArchiver.Open(const APath: TPath): IFolder;
begin
	if not TarEnabled then
		raise ETarError.CreateFmt(SFaultLoad, [Tar32Dll]);

	Result := TTarArchive.Create(APath)
end;

procedure TTarArchiver.Pack(const Archive, SourceDir: WideString; const Sources: array of WideString);
var
	ArchiveA, SourcePath, Cmd, ResponseFile: string;
	Buffer: array[0..4095] of Char;
begin
	try
		ArchiveA := Archive;
		SourcePath := IncludeTrailingPathDelimiter(SourceDir);

		{R}h}
		Cmd := '-c ';
		if UpCase(ArchiveA[Length(ArchiveA)]) = 'Z' then
			Cmd := Cmd + '-z '
		else if lstrcmpi(PChar(ArchiveA) + Length(ArchiveA) - 4, '.bz2') = 0 then
			Cmd := Cmd + '-B ';

		{ɂƊ{pX}
		Cmd := Cmd + '"' + ArchiveA +'" "' + SourcePath + '"';

		{Ώ}
		if Length(Sources) = 1 then
		begin
			if DirectoryExists(SourcePath + Sources[0]) then
				Cmd := Cmd + ' "' + Sources[0] +'\*"'
			else
				Cmd := Cmd + ' "' + Sources[0] +'"'
		end
		else
		begin
			ResponseFile := MakeResponseFile(SourcePath, Sources);
			Cmd := Cmd + ' @' + ExtractShortPathName(ResponseFile)
		end;

		if Tar(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
			raise ETarError.Create(Buffer);
	finally
		if ResponseFile <> '' then Windows.DeleteFile(PChar(ResponseFile));
	end
end;

procedure TTarArchiver.Unpack(const Archive, DestDir: WideString);
var
	Cmd: String;
	Buffer: array[0..4095] of Char;
begin
	Cmd := '-x "' + Archive + '" "' + IncludeTrailingPathDelimiter(DestDir) + '"';

	if Tar(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise ETarError.Create(Buffer);
end;

procedure Register;
begin
	(GetApplication as IExtensionManager).RegisterExtension(TTarArchiver.Create)
end;

initialization

	if GetInterfaceVersion = EOFileIntf.InterfaceVersion then
		Register;

end.
